<?php

namespace Drupal\config_override_warn;

use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Config\TypedConfigManagerInterface;
use Drupal\Core\Form\FormInterface;

/**
 * Contains logic for inspecting config forms and their overridden values.
 */
class FormOverrides {

  /**
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * @var \Drupal\Core\Config\TypedConfigManagerInterface
   */
  protected $typedConfigManager;

  /**
   * Constructs the FormOverrides service.
   *
   * @param ConfigFactoryInterface $config_factory
   *   The config factory.
   * @param TypedConfigManagerInterface $typed_config_manager
   *   The typed config manager.
   */
  public function __construct(ConfigFactoryInterface $config_factory, TypedConfigManagerInterface $typed_config_manager) {
    $this->configFactory = $config_factory;
    $this->typedConfigManager = $typed_config_manager;
  }

  /**
   * Get overrides for a form.
   *
   * @param FormInterface $form
   *   The form object.
   * @return array
   *   A nested array of overridden config values, keyed by config name, and subkeyed by config value name.
   *   The subvalue is an array with 'original' and 'override' values of the respective config property.
   */
  public function getFormOverrides(FormInterface $form) {
    $names = $this->getFormConfigNames($form);
    $overrides = [];
    foreach ($names as $name) {
      $overrides = NestedArray::mergeDeep($overrides, $this->getConfigOverrides($name));
    }
    return $overrides;
  }

  /**
   * Get the config names that correspond with a form.
   *
   * @param FormInterface $form
   *   The form object.
   * @return array
   *   An array of config names.
   */
  public function getFormConfigNames(FormInterface $form) {
    $overrides = [];
    if (method_exists($form, 'getEditableConfigNames')) {
      // Grr... this is a protected method on \Drupal\Core\Form\ConfigFormBaseTrait
      // @see https://www.drupal.org/project/drupal/issues/2095289
      $method = new \ReflectionMethod($form, 'getEditableConfigNames');
      $method->setAccessible(TRUE);
      $overrides = $method->invoke($form);
    }
    return $overrides;
  }

  /**
   * Get overrides for a config.
   *
   * @param $name
   *   The name of the configuration object
   *
   * @return array
   *   A nested array of overridden config values, keyed by config value name.
   *   The value is an array with 'original' and 'override' values of the respective config property.
   */
  public function getConfigOverrides($name) {
    $overrides = [];
    $config = $this->configFactory->get($name);
    if ($config->hasOverrides()) {
      $property = new \ReflectionProperty($config, 'settingsOverrides');
      $property->setAccessible(TRUE);
      if ($config_overrides = $property->getValue($config)) {
        $definition = $this->typedConfigManager->getDefinition($name);
        $keys = $this->getConfigKeys($config_overrides, $definition);
        foreach ($keys as $key) {
          $original_value = var_export($config->getOriginal($key, FALSE), TRUE);
          $override_value = var_export($config->get($key), TRUE);
          if ($original_value !== $override_value) {
            $overrides[$name][$key] = [
              'original' => $original_value,
              'override' => $override_value,
            ];
          }
        }
      }
    }
    return $overrides;
  }

  /**
   * Get all possible keys from a config object.
   *
   * @param array $values
   *   The root values from the config object.
   * @param array $definition
   *   The config definition for $values.
   * @param null $prefix
   *   Used for recursion of sub-keys.
   *
   * @return array
   *   An array of config keys.
   */
  protected function getConfigKeys(array $values, array $definition, $prefix = NULL) {
    $keys = [];
    foreach ($values as $key => $value) {
      if (is_array($value) && isset($definition['mapping'])) {
        $value_definition = NestedArray::getValue($definition['mapping'], explode('.', $key));
        if (isset($value_definition['type']) && $value_definition['type'] === 'mapping') {
          $keys = array_merge($keys, $this->getConfigKeys($value, $value_definition, $prefix . $key . '.'));
          continue;
        }
      }
      $keys[] = $prefix . $key;
    }
    return $keys;
  }

}
